# Copyright 2024 Agnostiq Inc.
"""Blueprint decorator."""

from functools import wraps

from covalent_blueprints.blueprints.utilities import (
    get_input_docs,
    get_kwargs,
    get_usage_example,
)


def blueprint(title: str = ""):
    """Decorator for Covalent Blueprint initializer functions. This decorator
    auto-sets the title and other metadata for the blueprint.

    Args:
        title: Optional title for the Blueprint. Defaults to the initializer
            function's name.
    """

    def blueprint_decorator(initializer_func):

        @wraps(initializer_func)
        def wrapper(*args, **kwargs):

            # Initialize the blueprint
            blueprint_instance = initializer_func(*args, **kwargs)
            # Set metadata
            blueprint_instance.title = title or initializer_func.__name__
            blueprint_instance.example = get_usage_example(initializer_func)
            blueprint_instance.inputs.docs = get_input_docs(initializer_func)
            blueprint_instance.inputs.kwargs = get_kwargs(initializer_func)

            return blueprint_instance

        return wrapper

    return blueprint_decorator



# Copyright 2024 Agnostiq Inc.
"""Logical representation of a Covalent script/module."""

import importlib
import sys
from enum import Enum
from functools import partial
from pathlib import Path
from types import ModuleType
from typing import Callable, Dict, List, Set, Tuple

import covalent_cloud as cc
from covalent._workflow.electron import Electron
from covalent._workflow.lattice import Lattice
from covalent_cloud.dispatch_management.interface_functions import dispatch
from covalent_cloud.function_serve.deployment import deploy
from covalent_cloud.function_serve.service_class import FunctionService
from covalent_cloud.shared.schemas.volume import Volume
from pkg_resources import Requirement
from pydantic.dataclasses import dataclass

from covalent_blueprints import __file__ as blueprints_file
from covalent_blueprints.logger import bp_log
from covalent_blueprints.reader.capture import (
    CapturedCloudExecutorArguments,
    CapturedCreateEnvArguments,
    CapturedDeployArguments,
    CapturedDispatchArguments,
    CapturedLatticeDeclaration,
    CapturedVolumeArguments,
    capture_cloud_calls,
)
from covalent_blueprints.reader.utilities import custom_build_graph

_BUILDING_BLOCK_TYPES = {Electron, Lattice, FunctionService}
_BLUEPRINTS_PKG = "covalent-blueprints"
try:
    _MIN_VERSION = (Path(blueprints_file).parent.parent / "VERSION").read_text().strip()
except FileNotFoundError:
    _MIN_VERSION = ""

_BLUEPRINTS_PKG_LATEST = (
    f"{_BLUEPRINTS_PKG}>={_MIN_VERSION}" if _MIN_VERSION else _BLUEPRINTS_PKG
)


class ScriptType(Enum):
    """Enumerates the type of Covalent scripts in terms of the main
    function call."""

    DISPATCH = "dispatch"
    DEPLOY = "deploy"


class BlueprintValidationError(Exception):
    """Base class for all blueprint exceptions."""


@dataclass(config={"arbitrary_types_allowed": True})
class CovalentScript:
    """Represents a Covalent workflow or service scripts (.py
    file)"""

    electrons: Dict[str, Callable]
    services: Dict[str, FunctionService]
    lattices: List[CapturedLatticeDeclaration]
    sublattices: Set[str]
    executors: List[CapturedCloudExecutorArguments]
    dispatches: List[CapturedDispatchArguments]
    dispatch_inputs: List[Tuple]
    deploys: List[CapturedDeployArguments]
    deploy_inputs: List[Tuple]
    volumes: List[CapturedVolumeArguments]
    environments: List[CapturedCreateEnvArguments]
    source: Path
    module: ModuleType

    def __post_init__(self):
        self._wrapper = None  # Callable - either `cc.dispatch` or `cc.deploy`
        self._script_type = None  # ScriptType - 'dispatch' or 'deploy'

        # Validate collection and enforce assumptions.

        if len(self.dispatches) > 1:
            raise BlueprintValidationError(
                "Can't convert covalent script with more than one `cc.dispatch` call."
            )
        if len(self.deploys) > 1:
            raise BlueprintValidationError(
                "Can't convert covalent script with more than one `cc.deploy` call. "
                "Use a lattice to deploy multiple services inside a recipe."
            )
        if len(self.dispatches) == 1 and len(self.deploys) == 1:
            raise BlueprintValidationError(
                "Can't convert covalent script with both `cc.dispatch` and "
                "`cc.deploy` calls. Consider deploying services inside the lattice."
            )
        if len(self.dispatches) == 0 and len(self.deploys) == 0:
            raise BlueprintValidationError(
                "Can't convert covalent script without a `cc.dispatch` "
                "nor `cc.deploy` call."
            )

        # Check that all envs used in the script are declared in the script
        for ex in self.executors:
            if ex.env == "default":
                continue
            if ex.env not in [env.name for env in self.environments]:
                raise BlueprintValidationError(
                    f"Environment '{ex.env}' is not created in the script. "
                )

        # Check that all env creation calls have the latest blueprints package
        for env in self.environments:
            if not any(Requirement(r).name == _BLUEPRINTS_PKG for r in env.pip):
                raise BlueprintValidationError(
                    f"Environment '{env.name}' should have '{_BLUEPRINTS_PKG}' "
                    f"as a pip dependency, e.g. '{_BLUEPRINTS_PKG_LATEST}'."
                )

    @property
    def type(self) -> ScriptType:
        """Whether the script is a dispatch or deploy script."""
        if self._script_type is None:
            if len(self.dispatches) == 1:
                self._script_type = ScriptType.DISPATCH
            elif len(self.deploys) == 1:
                self._script_type = ScriptType.DEPLOY
            else:
                raise self._indeterminate_type_error()

        return self._script_type

    @property
    def wrapper(self) -> Callable:
        """Either `cc.dispatch` or `cc.deploy`, depending on the
        script."""
        if self.type == ScriptType.DISPATCH:
            return partial(dispatch, volume=self.dispatches[0].volume)

        if self.type == ScriptType.DEPLOY:
            return partial(deploy, volume=self.deploys[0].volume)

        raise self._indeterminate_type_error()

    @property
    def core_function_inputs(self) -> Tuple:
        """The inputs to the main lattice or main function
        service."""
        if self.type == ScriptType.DISPATCH:
            return self.dispatch_inputs[0]

        if self.type == ScriptType.DEPLOY:
            return self.deploy_inputs[0]

        raise self._indeterminate_type_error()

    @property
    def volume(self) -> Volume:
        """The volume attached at dispatch or deploy."""
        if self.type == ScriptType.DISPATCH:
            return self.dispatches[0].volume

        if self.type == ScriptType.DEPLOY:
            return self.deploys[0].volume

        raise self._indeterminate_type_error()

    def get_core_function(self) -> Callable:
        """Either the main lattice or the main function service,
        including post-build updates."""

        bp_log.debug(
            "Getting core function for script '%s' ('%s')",
            self.source.name,
            self.type.value,
        )

        if self.type == ScriptType.DISPATCH:
            lattice_func = self.dispatches[0].lattice_func
            lattice_func_name = lattice_func.__name__
            return self._get_captured_lattice(lattice_func_name).lattice_obj

        if self.type == ScriptType.DEPLOY:
            function_service = self.deploys[0].function_service
            return self.services[function_service.func_name]

        raise self._indeterminate_type_error()

    def get_cloud_executable(self, *args, **kwargs) -> Callable:
        """Either `cc.dispatch(core_lattice)` or
        `cc.deploy(core_service)`"""
        func = self.get_core_function()

        if self.type == ScriptType.DISPATCH:
            # The top level lattice is built locally during dispatch.
            # The following makes sure overloaded resources are applied.
            bp_log.debug(
                "Running custom_build_graph locally on lattice '%s'", func.__name__
            )
            custom_build_graph(func, self.electrons, *args, **kwargs)  # type: ignore

        return self.wrapper(func)

    def create_envs(self, wait: bool = True) -> None:
        """Create the environments used in the script."""
        for env in self.environments:
            kwargs = env.__dict__.copy()
            kwargs["wait"] = wait
            cc.create_env(**kwargs)

    def _get_captured_lattice(self, lattice_name: str) -> CapturedLatticeDeclaration:
        """Get the captured lattice object by name."""
        try:
            return next(
                cl
                for cl in self.lattices
                if cl._func.__name__ == lattice_name  # pylint: disable=protected-access
            )
        except StopIteration as e:
            raise ValueError(
                f"No captured declaration for lattice '{lattice_name}'"
            ) from e

    def _indeterminate_type_error(self) -> RuntimeError:
        return RuntimeError("Can't determine whether to dispatch or deploy script.")

    @staticmethod
    def _read(module) -> Tuple[dict, dict]:
        """Extract any electrons or services from the module.

        Must call inside `capture_cloud_calls` context to avoid
        initializing real cloud resources.
        """

        electrons = {}
        services = {}

        # Collect covalent objects.
        for obj_name, obj in module.__dict__.items():

            # Check each object.
            if not callable(obj):
                continue
            if (type_ := _get_callable_type(obj)) not in _BUILDING_BLOCK_TYPES:
                continue

            # Sort each object.
            if type_ is Electron:
                electrons[obj_name] = obj

            elif type_ is FunctionService:
                services[obj_name] = obj

        return electrons, services

    @staticmethod
    def patched_import_module(module_path: Path) -> Tuple[ModuleType, dict]:
        """Import module with cloud calls disabled and Covalent
        object captured."""
        sys_path = sys.path.copy()
        sys.path.append(str(module_path.parent))

        module_name = module_path.with_suffix("").name
        with capture_cloud_calls() as script_data:
            if module_name not in sys.modules:
                module = importlib.import_module(module_name)
            else:
                module = importlib.reload(sys.modules[module_name])

        sys.path = sys_path

        return module, script_data

    @classmethod
    def from_module(cls, module_path: Path) -> "CovalentScript":
        """Convert a module (covalent script) into a `CovalentScript`
        object."""

        module, script_data = cls.patched_import_module(module_path)
        electrons, services = CovalentScript._read(module)

        return cls(
            electrons=electrons,
            services=services,
            source=module_path.absolute(),
            module=module,
            **script_data,
        )


def _get_callable_type(func):
    return Electron if hasattr(func, "electron_object") else type(func)



# Copyright 2024 Agnostiq Inc.
"""Source file for testing purposes only.

This script implements a chatbot backend service.
"""
import covalent_cloud as cc

# cc.save_api_key("you-api-key")

ENV_NAME = "llm-peft-lite"

cc.create_env(
    name=ENV_NAME,
    pip=["accelerate", "sentencepiece", "transformers", "covalent-blueprints"],
    wait=True,
)

gpu_executor = cc.CloudExecutor(
    env=ENV_NAME,
    num_cpus=24,
    memory="54 GB",
    time_limit="15 days",
    num_gpus=1,
    gpu_type=cc.cloud_executor.GPU_TYPE.A100,
)


@cc.service(executor=gpu_executor, name="LLM Chatbot Server")
def chatbot_backend(model_path: str, device_map="auto"):
    """Create a Llama2 chatbot server."""

    import torch
    from transformers import AutoModelForCausalLM, AutoTokenizer, pipeline

    model = AutoModelForCausalLM.from_pretrained(
        model_path,
        device_map=device_map,
        torch_dtype=torch.float16,
        do_sample=True,
    )
    tokenizer = AutoTokenizer.from_pretrained(model_path)
    pipe = pipeline(task="text-generation", model=model, tokenizer=tokenizer)
    return {"pipe": pipe}


@chatbot_backend.endpoint("/generate", name="Generate Response")
def generate(pipe, prompt, max_new_tokens=50):
    """Generate a response to a prompt."""
    output = pipe(
        prompt,
        max_new_tokens=max_new_tokens,
        do_sample=True,
        truncation=True,
        temperature=0.9,
    )
    gen_text = output[0]["generated_text"]
    return gen_text


@chatbot_backend.endpoint("/stream", name="Stream Response", streaming=True)
def generate_stream(pipe, prompt, max_new_tokens=200):
    """Generate a response to a prompt, streaming tokens."""

    import torch

    def _starts_with_space(tokenizer, token_id):
        token = tokenizer.convert_ids_to_tokens(token_id)
        return token.startswith("▁")

    model = pipe.model
    tokenizer = pipe.tokenizer
    _input = tokenizer(prompt, return_tensors="pt").to("cuda")

    for output_length in range(max_new_tokens):
        # Generate next token
        output = model.generate(
            **_input,
            max_new_tokens=1,
            do_sample=True,
            temperature=0.9,
            pad_token_id=tokenizer.eos_token_id,
        )
        # Check for stopping condition
        current_token_id = output[0][-1]
        if current_token_id == tokenizer.eos_token_id:
            break
        # Decode token
        current_token = tokenizer.decode(current_token_id, skip_special_tokens=True)
        if _starts_with_space(tokenizer, current_token_id.item()) and output_length > 1:
            current_token = " " + current_token

        yield current_token

        # Update input for next iteration.
        # Output grows in size with each iteration.
        _input = {
            "input_ids": output.to("cuda"),
            "attention_mask": torch.ones(1, len(output[0])).to("cuda"),
        }


info = cc.deploy(chatbot_backend)(model_path="NousResearch/Llama-2-7b-chat-hf")
info = cc.get_deployment(info.function_id, wait=True)
print(info)
print(info.address)



# Copyright 2024 Agnostiq Inc.
"""Templates for Covalent scripts that run a workflow that executes a
single electron and deploys a single service."""

from typing import Type

from covalent_cloud.cloud_executor.cloud_executor import CloudExecutor

from covalent_blueprints.blueprints.blueprints import CovalentBlueprint
from covalent_blueprints.blueprints.executor_map import ExecutorMap
from covalent_blueprints.reader.script import CovalentScript

# pylint: disable=unnecessary-dunder-call


class ServiceWorkflowExecutorMap(ExecutorMap):
    """Thin wrapper mostly for convenient auto-completions."""

    def __init__(self, script: CovalentScript):
        super().__init__(script)
        self._electron_key = ""
        self._service_key = ""

    @property
    def task_executor(self) -> CloudExecutor:
        """Executor for the electron deployment."""
        if not self._electron_key:
            raise ValueError("Please set the executor map's electron key.")
        return self.__getitem__(self._electron_key)

    @task_executor.setter
    def task_executor(self, executor: CloudExecutor) -> None:
        if not self._electron_key:
            raise ValueError("Please set the executor map's electron key.")
        self.__setitem__(self._electron_key, executor)

    @property
    def service_executor(self) -> CloudExecutor:
        """Executor for the service deployment."""
        if not self._service_key:
            raise ValueError("Please set the executor map's service key.")
        return self.__getitem__(self._service_key)

    @service_executor.setter
    def service_executor(self, executor: CloudExecutor) -> None:
        if not self._service_key:
            raise ValueError("Please set the executor map's service key.")
        self.__setitem__(self._service_key, executor)

    def set_executor_key(self, electron_key: str, service_key: str) -> None:
        """Set the executor keys for the service workflow
        deployment."""
        self._electron_key = electron_key
        self._service_key = service_key


class ServiceWorkflowBlueprint(CovalentBlueprint):
    """A runnable blueprint that represents a workflow that executes
    one electron and deploys one service."""

    executor_map_type: Type = ServiceWorkflowExecutorMap

    @property
    def executors(self) -> ServiceWorkflowExecutorMap:
        return self._executors



# Copyright 2024 Agnostiq Inc.
"""Arguments for Covalent blueprints."""
from pprint import pformat
from typing import Dict, Optional

from covalent_blueprints.reader.script import CovalentScript


class BlueprintInputs:
    """Provides arguments interface for Covalent blueprints."""

    def __init__(self, script: CovalentScript):
        self._docs: Dict[str, Optional[str]] = {}

        # Arguments for the wrapped lattice or service
        self._core_args, self._core_kwargs = script.core_function_inputs

        if not self._core_args == ():
            raise ValueError(
                "Please avoid positional arguments in the core function "
                "(lattice or service) wrapped by the blueprint."
            )

        # Keyword arguments for the blueprint initializer.
        # These are set by the @blueprint decorator.
        self._kwargs: dict = {}

    @property
    def core_kwargs(self):
        return self._core_kwargs

    @core_kwargs.setter
    def core_kwargs(self, value: dict):
        if isinstance(value, dict):
            self._core_kwargs = value
        else:
            raise ValueError("core_kwargs must be a dict")

    @property
    def kwargs(self) -> dict:
        """Default keyword arguments for the blueprint's core
        function."""
        return self._kwargs

    @kwargs.setter
    def kwargs(self, value: dict) -> None:
        if isinstance(value, dict):
            self._kwargs = value
        else:
            raise ValueError("kwargs must be a dict")

    @property
    def docs(self) -> Dict[str, Optional[str]]:
        """Documentation for the arguments."""
        return self._docs.copy()

    @docs.setter
    def docs(self, value: dict) -> None:
        self._docs = value

    def to_dict(self):
        """Return the arguments as a dictionary."""
        return {"kwargs": self.kwargs.copy(), "docs": self.docs.copy()}

    def override_core_defaults(self, kwargs) -> dict:
        """Override the default arguments with the provided
        kwargs."""
        new_kwargs = self._core_kwargs.copy()
        new_kwargs.update(**kwargs)

        return new_kwargs

    def __getitem__(self, key):
        if key == "args":
            return self._core_args
        if key == "kwargs":
            return self.kwargs
        raise KeyError(f"Invalid key '{key}'")

    def __repr__(self):
        return f"BlueprintInputs({self.kwargs})"

    def __str__(self):
        return pformat(self.to_dict())



# Copyright 2024 Agnostiq Inc.
"""Utilities for processing Covalent scripts."""

from typing import Callable, Dict

import covalent as ct
from covalent._workflow.lattice import Lattice

from covalent_blueprints.logger import bp_log


def custom_build_graph(
    lattice_obj: Lattice, electrons: Dict[str, Callable], *args, **kwargs
) -> None:
    """Override the usual `build_graph` method of a lattice to update
    metadata.

    Args:
        lattice_: The lattice object to update.
    """
    # pylint: disable=protected-access
    bp_log.debug(
        "Running native build_graph on lattice\n\n%s",
        lattice_obj.workflow_function_string,
    )

    # Need task_packing "context" like in `cc.dispatch`.
    old_task_packing = ct.get_config("sdk.task_packing")
    ct.set_config("sdk.task_packing", "true")

    # Calling via class avoids recursion errors in sublattice case.
    Lattice.build_graph(lattice_obj, *args, **kwargs)

    ct.set_config("sdk.task_packing", old_task_packing)

    # Map of electron names to updated metadata.
    electron_metadata_dict = {
        name: electron.electron_object.metadata.copy()  # type: ignore
        for name, electron in electrons.items()
    }

    # Map of task group IDs to updated metadata.
    new_executor_data_by_group = {}

    # Identify electron task groups and prepare metadata for them.
    for node_id in lattice_obj.transport_graph._graph.nodes:
        node_dict = lattice_obj.transport_graph._graph.nodes[node_id]
        name = node_dict["name"]

        if new_metadata := electron_metadata_dict.get(name):

            # Prepare metadata for the electron's task group.
            task_group_id = node_dict["task_group_id"]
            new_executor_data_by_group[task_group_id] = new_metadata["executor_data"]

            # Shorten loop to avoid redundant iterations.
            del electron_metadata_dict[name]

        # Exit once all electrons have been processed.
        if not electron_metadata_dict:
            break

    bp_log.debug(
        "Updating executor data for task groups: %s",
        tuple(new_executor_data_by_group.keys()),
    )

    # Loop again to set metadata for task groups.
    for node_id in lattice_obj.transport_graph._graph.nodes:
        node_dict = lattice_obj.transport_graph._graph.nodes[node_id]
        task_group_id = node_dict["task_group_id"]

        if task_group_id in new_executor_data_by_group:

            # Set the electron's executor for tasks in its task group.
            executor_data = new_executor_data_by_group[task_group_id]
            bp_log.debug(
                "Setting executor data for task group %s:\n%s\n\n",
                task_group_id,
                executor_data,
            )
            node_dict["metadata"]["executor_data"] = executor_data

    bp_log.debug(
        "Nullifying build_graph on lattice %s ('%s) after custom_build_graph",
        lattice_obj,
        lattice_obj.__name__,
    )
    setattr(lattice_obj, "build_graph", lambda *_, **__: None)



# Copyright 2024 Agnostiq Inc.
"""Implements a model for blueprint summaries."""

from pathlib import Path
from typing import Any, Dict, List, Optional, Union

from pydantic import BaseModel, ConfigDict, field_validator


class BlueprintSummary(BaseModel):
    """Detailed summary of a Covalent blueprint."""

    name: str
    title: str
    description: Union[str, None]
    type: str
    inputs: Dict[str, Any]
    executors: Dict[str, Any]
    environments: List[Dict[str, Any]]
    volumes: List[str]
    example: str
    source_file: Optional[str] = None
    source: Optional[str] = None

    model_config = ConfigDict(extra="forbid")

    @field_validator("name")
    @classmethod
    def _clean_up_name(cls, value):
        return value.lstrip("_")

    @field_validator("description")
    @classmethod
    def _clean_up_description(cls, value):
        value_ = value.strip() if value else ""
        value_ = value_.replace("\n", " ")
        return value_

    @field_validator("executors")
    @classmethod
    def _clean_up_executors(cls, value):
        remove_fields = ["settings", "validate_environment", "volume_id"]
        for target, executor_dict in value.items():
            value[target] = {
                k: v for k, v in executor_dict.items() if k not in remove_fields
            }
        return value

    @field_validator("environments")
    @classmethod
    def _clean_up_environments(cls, value):
        remove_fields = ["timeout", "wait"]
        new_envs = []
        for env in value:
            env = {k: v for k, v in env.items() if k not in remove_fields}
            new_envs.append(env)
        return new_envs

    def get_source(self):
        """Read the source script from the source file."""
        if not Path(self.source_file).is_file():
            raise FileNotFoundError(f"File not found: {self.source_file}")

        # Read source script.
        self.source = self.source_file.read_text()



# Copyright 2024 Agnostiq Inc.
""""Tests for the often-used `get_blueprint` function."""
from pathlib import Path

import pytest

from covalent_blueprints import get_blueprint
from covalent_blueprints.blueprints.blueprints import CovalentBlueprint
from covalent_blueprints.blueprints.templates import (
    ServiceWorkflowBlueprint,
    SingleServiceBlueprint,
)

CWD = Path(__file__).parent


@pytest.mark.parametrize(
    "module_path,_cls",
    [
        (CWD / "samples/_scripts/_deploy_chatbot.py", SingleServiceBlueprint),
    ],
)
def test_get_from_source_script(module_path, _cls):
    """Test getting a blueprint from a source .py file."""

    bp = get_blueprint(module_path=module_path, _cls=_cls)
    assert isinstance(bp, _cls)


@pytest.mark.parametrize(
    "module_path,_cls",
    [
        (CWD / "samples/_notebooks/_basic_notebook.ipynb", CovalentBlueprint),
        (
            CWD / "samples/_notebooks/_image_generator_service.ipynb",
            ServiceWorkflowBlueprint,
        ),
    ],
)
def test_get_from_source_notebook(module_path, _cls):
    """Test getting a blueprint from a source .ipynb file."""

    bp = get_blueprint(module_path=module_path, _cls=_cls)
    assert isinstance(bp, _cls)



# Copyright 2024 Agnostiq Inc.
"""A collection of reusable blueprint types for creating installed
blueprints."""

from .single_service_deployment import SingleServiceBlueprint
from .single_service_workflow import ServiceWorkflowBlueprint



# Copyright 2024 Agnostiq Inc.
"""Module for capturing essential elements from Covalent scripts."""

import contextlib
from datetime import timedelta
from pprint import pformat
from typing import Any, Callable, Dict, List, Optional, Union
from unittest.mock import MagicMock, patch

import covalent_cloud as cc
from covalent._workflow.lattice import Lattice
from covalent_cloud.function_serve.service_class import FunctionService
from covalent_cloud.shared.classes.settings import Settings, settings
from covalent_cloud.shared.schemas.volume import Volume
from pydantic.dataclasses import Field, dataclass

from covalent_blueprints.logger import bp_log

CreateEnvCondaArgs = Optional[Union[str, List[str], Dict[str, List[str]]]]


@dataclass
class CapturedCreateEnvArguments:
    """Arguments captured from `cc.create_env()` calls."""

    name: str
    pip: Optional[Union[str, List[str]]] = Field(default_factory=list)
    conda: CreateEnvCondaArgs = Field(default_factory=list)
    variables: Optional[List] = None
    settings: Optional[Settings] = Field(default_factory=lambda: settings)
    wait: Optional[bool] = False
    timeout: Optional[int] = 1800
    base_image: Optional[str] = None


@dataclass
class CapturedVolumeArguments:
    """Arguments captured from `cc.volume()` calls."""

    name: str
    vtype: Optional[str] = "OBJECT_STORAGE"
    settings: Optional[Settings] = Field(default_factory=lambda: settings)


@dataclass(config={"arbitrary_types_allowed": True})
class CapturedLatticeDeclaration:
    """Arguments captures from `ct.lattice()` calls."""

    lattice_obj: Lattice
    _func: Callable
    backend: Optional[str] = None
    executor: Optional[Any] = None
    workflow_executor: Optional[Any] = None
    deps_bash: Optional[Any] = None
    deps_pip: Optional[Any] = None
    call_before: Optional[Any] = None
    call_after: Optional[Any] = None
    triggers: Optional[Any] = None

    def get_kwargs(self) -> dict:
        """Return kwargs for lattice decorator."""
        d = self.__dict__.copy()
        d.pop("lattice_obj")
        return d


@dataclass(config={"arbitrary_types_allowed": True})
class CapturedDispatchArguments:
    """Arguments captured from `cc.dispatch()` calls."""

    lattice_func: Callable
    settings: Settings = Field(default_factory=lambda: settings)
    volume: Union[Volume, None] = None


@dataclass(config={"arbitrary_types_allowed": True})
class CapturedDeployArguments:
    """Arguments captured from `cc.deploy()` calls."""

    function_service: FunctionService
    volume: Optional[Volume] = None


@dataclass
class CapturedCloudExecutorArguments:
    """Arguments captured from `cc.CloudExecutor()`
    instantiations."""

    num_cpus: int = 1
    memory: Union[int, str] = 1024
    num_gpus: int = 0
    gpu_type: Union[str, cc.cloud_executor.GPU_TYPE] = ""
    env: str = "default"
    time_limit: Union[int, timedelta, str] = 60 * 30
    volume_id: Optional[int] = None
    settings: Dict = Field(default_factory=settings.model_dump)
    validate_environment: bool = True


@contextlib.contextmanager
def capture_cloud_calls():
    """Context designed to capture Covalent function calls and
    objects instantiations, when a script is imported inside it."""

    # pylint: disable=import-outside-toplevel

    # Aliases for real Covalent functions to avoid patches.
    from covalent._workflow.electron import electron as ct_electron
    from covalent._workflow.lattice import Lattice as ct_Lattice
    from covalent._workflow.lattice import lattice as ct_lattice
    from covalent_cloud.cloud_executor.cloud_executor import (
        CloudExecutor as cc_CloudExecutor,
    )
    from covalent_cloud.volume.volume import volume as cc_volume

    class _Result:
        """Mock result object to patch typical manipulations."""

        # pylint: disable=missing-function-docstring
        @property
        def result(self):
            return self

        @property
        def value(self):
            return None

        def load(self):
            return self

    class _Deployment:
        """Mock deployment object to patch typical manipulations."""

        # pylint: disable=too-few-public-methods

        def __getattribute__(self, name: str) -> None:
            return None

    script_data = {
        "environments": [],
        "volumes": [],
        "lattices": [],
        "sublattices": set(),
        "dispatches": [],
        "deploys": [],
        "dispatch_inputs": [],
        "deploy_inputs": [],
        "executors": [],
    }

    def _capture_create_env(*args, **kwargs):
        captured_env = CapturedCreateEnvArguments(*args, **kwargs)
        bp_log.debug("Captured environment:\n%s", pformat(captured_env.__dict__))
        script_data["environments"].append(captured_env)

    def _capture_volume(*args, **kwargs):
        captured_volume = CapturedVolumeArguments(*args, **kwargs)
        bp_log.debug("Captured volume:\n%s", pformat(captured_volume.__dict__))
        script_data["volumes"].append(captured_volume)
        # Actually create the volume here.
        return cc_volume(*args, **kwargs)

    def _capture_electron_declaration(_func=None, **kwargs):

        def _capture_electron_wrapper(_func=None):
            bp_log.debug("Captured electron wrapper processing: %s", _func)
            if isinstance(_func, ct_Lattice) or (
                hasattr(_func, __name__)
                and _func.__name__ == "_capture_lattice_wrapper"
            ):
                script_data["sublattices"].add(_func.__name__)
                bp_log.debug("Registered sublattice: %s", _func.__name__)

            return ct_electron(_func, **kwargs)

        if _func is None:
            return _capture_electron_wrapper
        return _capture_electron_wrapper(_func)

    def _capture_lattice_declaration(_func=None, **kwargs):

        def _capture_lattice_wrapper(_func=None):
            bp_log.debug("Captured lattice wrapper processing: %s", _func)

            lattice_obj = ct_lattice(_func, **kwargs)
            captured_lattice = CapturedLatticeDeclaration(lattice_obj, _func, **kwargs)
            script_data["lattices"].append(captured_lattice)

            return lattice_obj

        if _func is None:
            return _capture_lattice_wrapper
        return _capture_lattice_wrapper(_func)

    def _capture_dispatch(*args, **kwargs):
        captured_dispatch = CapturedDispatchArguments(*args, **kwargs)
        bp_log.debug("Captured dispatch:\n%s", pformat(captured_dispatch.__dict__))
        script_data["dispatches"].append(captured_dispatch)
        return lambda *a, **k: script_data["dispatch_inputs"].append((a, k))

    def _capture_deploy_inputs(*a, **k):
        bp_log.debug("Captured deploy inputs: %s", (a, k))
        script_data["deploy_inputs"].append((a, k))
        return _Deployment()

    def _capture_deploy(*args, **kwargs):
        captured_deploy_args = CapturedDeployArguments(*args, **kwargs)
        bp_log.debug(
            "Captured deploy call:\n%s", pformat(captured_deploy_args.__dict__)
        )
        script_data["deploys"].append(captured_deploy_args)
        return _capture_deploy_inputs

    def _capture_cloud_executor(*args, **kwargs):
        captured_executor = CapturedCloudExecutorArguments(*args, **kwargs)
        bp_log.debug(
            "Captured cloud executor:\n%s", pformat(captured_executor.__dict__)
        )
        script_data["executors"].append(captured_executor)
        kwargs["validate_environment"] = False
        return cc_CloudExecutor(*args, **kwargs)

    def _error_on_os_dispatch(*_, **__):
        raise ValueError(
            "Detected use of `ct.dispatch`. Please replace with `cc.dispatch`"
        )

    with patch("covalent_cloud.create_env", _capture_create_env), patch(
        "covalent_cloud.volume", _capture_volume
    ), patch("covalent.electron", _capture_electron_declaration), patch(
        "covalent.lattice", _capture_lattice_declaration
    ), patch(
        "covalent_cloud.dispatch", _capture_dispatch
    ), patch(
        "covalent_cloud.deploy", _capture_deploy
    ), patch(
        "covalent_cloud.CloudExecutor", _capture_cloud_executor
    ), patch(
        "covalent_cloud.get_deployment", lambda *_, **__: _Deployment()
    ), patch(
        "covalent_cloud.get_result", lambda *_, **__: _Result()
    ), patch(
        "covalent_cloud.save_api_key", lambda *_, **__: None
    ), patch(
        "builtins.print", MagicMock()
    ), patch(
        "covalent.dispatch", _error_on_os_dispatch
    ):
        yield script_data



# Copyright 2024 Agnostiq Inc.
"""High-level utilities for the Covalent blueprints."""
import inspect
import json
import pydoc
import re
import shutil
import subprocess
import tempfile
import textwrap
import warnings
from pathlib import Path
from subprocess import CalledProcessError
from typing import Any, Callable, Dict, Optional, Type, Union

from covalent_blueprints.blueprints.blueprints import CovalentBlueprint
from covalent_blueprints.logger import bp_log
from covalent_blueprints.reader.script import CovalentScript

BLUEPRINTS_CACHE_DIR = Path.home() / ".cache/covalent/blueprints"
REGISTRY_FILE = BLUEPRINTS_CACHE_DIR / "install_dirs.json"

# Parent path for these is the `covalent_blueprints` package.
FIND_SCRIPTS_PATTERN = r"**/_src/_*.py"
FIND_NOTEBOOKS_PATTERN = r"**/_src/_*.ipynb"

if not REGISTRY_FILE.exists():
    REGISTRY_FILE.parent.mkdir(parents=True, exist_ok=True)
    with open(REGISTRY_FILE, "w", encoding="utf-8") as registry_file:
        # Create an empty registry file.
        json.dump({}, registry_file)


def get_blueprint(
    module_path: Union[Path, str],
    skip_shell_commands: bool = True,
    check_installed: bool = True,
    _cls: Type[CovalentBlueprint] = CovalentBlueprint,
) -> Any:
    """Load the Python module at the given path and return a
    CovalentBlueprint object.

    Args:
        module_path: Path to the Python module or Jupyter notebook.
        skip_shell_commands: Skip shell commands inside Jupyter notebooks.
            Defaults to True.
        check_installed: Find installed blueprint if module path does not exist.
            Defaults to True.

    Raises:
        FileNotFoundError: If the module path does not exist nor is an installed
            blueprint.
        ValueError: If the module path extension is neither '.py' nor '.ipynb'.

    Returns:
        A blueprint object.
    """

    if not issubclass(_cls, CovalentBlueprint):
        raise TypeError(
            f"Expected a subclass of CovalentBlueprint, got {_cls.__name__}"
        )

    bp_log.debug(
        "Getting blueprints from module path: %s (type: %s)", module_path, _cls.__name__
    )

    module_path = Path(module_path)

    if not module_path.is_file():
        if not check_installed:
            raise FileNotFoundError(f"File not found: {module_path}")

        # Try to find an installed blueprint.
        if installed_blueprint := find_blueprint(str(module_path), _cls):
            return installed_blueprint
        raise ValueError(_format_installed_not_found_error(module_path))

    # File exists. It must be a Python script or Jupyter notebook.
    if module_path.suffix not in {".py", ".ipynb"}:
        raise ValueError("Only Python scripts or Jupyter notebooks are supported")

    # Handle Jupyter notebooks.
    if module_path.suffix == ".ipynb":
        module_path = _py_file_from_notebook(module_path, skip_shell_commands)

    return _load_blueprint_from_module(module_path, _cls)


def find_blueprint(
    name: str,
    _cls: Type[CovalentBlueprint] = CovalentBlueprint,
) -> Optional[Any]:
    """Load the installed blueprint with the given name and return a
    CovalentBlueprint object.

    Args:
        name: Name of the installed blueprint.

    Returns:
        A blueprint object or None if the blueprint is not found.
    """

    installed_blueprints = _find_installed_blueprints()
    module_path = installed_blueprints.get(name)

    if module_path is None:
        bp_log.debug("Blueprint '%s' not found", name)
        return None

    if module_path.suffix == ".ipynb":
        module_path = _py_file_from_notebook(module_path, skip_shell_commands=True)

    return _load_blueprint_from_module(module_path, _cls)


def _load_blueprint_from_module(
    module_path: Path,
    _cls: Type[CovalentBlueprint] = CovalentBlueprint,
) -> CovalentBlueprint:
    """Helper to load a blueprint from a path to a Python module."""

    script = CovalentScript.from_module(module_path)
    executor_map = _cls.create_executor_map(script)
    blueprint = _cls(module_path.stem, script, executor_map)

    return blueprint


def _py_file_from_notebook(
    notebook_path: Path, skip_shell_commands: bool = True
) -> Path:
    """Helper to convert Jupyter notebook into Python file."""

    try:
        subprocess.run(
            ["which", "jupyter-notebook"],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            check=True,
        )
    except CalledProcessError as e:
        raise RuntimeError(
            "Jupyter-notebook is required to convert notebooks to python scripts. "
            "Please retry after installing it with `pip install jupyterlab`."
        ) from e

    BLUEPRINTS_CACHE_DIR.mkdir(parents=True, exist_ok=True)
    py_file_name = f"{notebook_path.with_suffix('').name}.py"

    with open(notebook_path, "r", encoding="utf-8") as file:
        notebook_dict = json.load(file)

    keep_cells = []
    for cell in notebook_dict["cells"]:
        if not cell["cell_type"] == "code":
            # Skip non-code cells.
            continue

        if (
            (metadata := cell.get("metadata"))
            and "vscode" in metadata
            and metadata["vscode"].get("languageId")
        ):
            # Skip non-python cells.
            continue

        keep_source = []
        for line in cell.get("source", []):
            if skip_shell_commands and line.startswith("!"):
                # Skip shell commands.
                continue

            keep_source.append(line)

        cell["source"] = keep_source
        keep_cells.append(cell)

    notebook_dict["cells"] = keep_cells

    # Write the sanitized notebook to a temp file,
    # then convert temp notebook to script.
    with tempfile.NamedTemporaryFile(mode="w", suffix=".ipynb") as temp_notebook:
        json.dump(notebook_dict, temp_notebook, indent=2)
        temp_notebook.flush()

        # Will save converted script to cache directory.
        py_file_path = BLUEPRINTS_CACHE_DIR / py_file_name

        try:
            # Convert notebook to script.
            cmd = (
                f"jupyter nbconvert --to script {temp_notebook.name} "
                f"--log-level WARN --output {py_file_path.with_suffix('')}"
            )
            subprocess.run(cmd.split(), check=True)
        except CalledProcessError as e:
            raise RuntimeError(
                f"Failed to convert jupyter notebook '{notebook_path!s}' to "
                "python script. "
            ) from e

    return py_file_path


def register_blueprints_dir(
    name: str,
    install_dir: Union[str, Path],
    overwrite: bool = False,
) -> None:
    """Register a directory that contains installed blueprints in
    it's sub- directories. The resolved, absolute path of the
    directory is stored in the registry.

    Registering a directory means adding it to the `instal_dirs.json` file in
    the blueprints cache directory located at
    '$HOME/.cache/covalent/blueprints'.

    Args:
        name: Name of the blueprints collection.
        install_dir: String or path object of the directory containing installed
            blueprints.
        overwrite: Overwrite the existing entry if it exists. Defaults to False.
    """

    if not BLUEPRINTS_CACHE_DIR.is_dir():
        BLUEPRINTS_CACHE_DIR.mkdir(parents=True, exist_ok=True)

    install_dir = Path(install_dir).resolve().absolute()

    # Check if the install directory exists.
    if not install_dir.is_dir():
        raise FileNotFoundError(
            f"Blueprints location '{install_dir}' does not exist or is not a directory"
        )

    # Load the registry.
    with open(REGISTRY_FILE, "r", encoding="utf-8") as file:
        registry_dict = json.load(file)

    # Filter out registered directories that no longer exist.
    keep_keys = []
    for name_, install_dir_ in registry_dict.items():
        if Path(install_dir_).is_dir():
            keep_keys.append(name_)
        else:
            warnings.warn(
                f"Install directory '{install_dir_}' registered under name '{name_}' "
                "no longer exists. Removing from registry."
            )
    registry_dict = {k: v for k, v in registry_dict.items() if k in keep_keys}

    # Check if the name already exists in the registry.
    if not overwrite and name in registry_dict:
        warnings.warn(
            f"Name '{name}' already exists in the registry. "
            "Pass `overwrite=True` to overwrite the existing entry."
        )
        return

    # Check if the install directory already exists in the registry.
    if not overwrite and str(install_dir) in registry_dict.values():
        warnings.warn(
            f"Install directory '{install_dir!s}' already exists in the registry "
            f"under name '{name}'."
        )
        if str(install_dir.absolute()) == registry_dict.get(name):
            # Already registered.
            return

    # Update the registry
    registry_dict[name] = str(install_dir)

    # Update the registry file.
    with open(REGISTRY_FILE, "w", encoding="utf-8") as file:
        json.dump(registry_dict, file, indent=2)


def _find_installed_blueprints() -> Dict[str, Path]:
    """Find all the blueprints in the covalent-blueprints package."""

    # Load the registry.
    with open(REGISTRY_FILE, "r", encoding="utf-8") as file:
        registry_dict = json.load(file)

    # Discover scripts and notebooks
    installed_blueprints = {}
    for name, install_dir in registry_dict.items():
        install_dir = Path(install_dir)
        blueprints_list = []
        blueprints_list.extend(install_dir.glob(FIND_SCRIPTS_PATTERN))
        blueprints_list.extend(install_dir.glob(FIND_NOTEBOOKS_PATTERN))

        bp_log.info("Blueprints found in '%s': %s", name, blueprints_list)
        if not blueprints_list:
            # Skip empty.
            continue

        installed_blueprints[name] = blueprints_list

    discovered_blueprints = {}
    for name, blueprints_list in installed_blueprints.items():
        for blueprint_file in blueprints_list:

            # Remove leading underscore for registered name.
            # Path conversion here is just to remove suffix.
            file_basename = str(Path(blueprint_file.name.lstrip("_")).with_suffix(""))

            file_name = "/".join([name, file_basename])
            discovered_blueprints[file_name] = blueprint_file.absolute()

    return discovered_blueprints


def _format_installed_not_found_error(module_path: Union[str, Path]) -> str:
    """Formats the error message that lists installed blueprints when
    a blueprint is not found."""
    installed_blueprints = _find_installed_blueprints()
    installed_ = []
    for bp_name, path in installed_blueprints.items():
        bp_name = f"'{bp_name}'"
        installed_.append(f"{bp_name}\n    @ {path}\n")

    installed_ = sorted(installed_)
    installed_str = "\n".join(installed_)
    return (
        f"Blueprint '{module_path!s}' not found. Installed blueprints:\n\n"
        f"{installed_str if installed_str.strip() else '[None]'}"
    )


def get_usage_example(func: Callable) -> str:
    """Get the usage example from the docstring of a function with an
    'Example:' section.

    Args:
        func: A function object.

    Returns:
        Usage example substring, or the empty string if no example is
        found.
    """
    doc_heading_pattern = re.compile(r"^(\w+:)($|\n)")
    docstring_lines = pydoc.getdoc(func).splitlines()

    # Find the example section.
    i = 0
    line = ""
    while i < len(docstring_lines):
        line = docstring_lines[i]
        if doc_heading_pattern.match(line) and "Example" in line:
            break
        i += 1

    # If no example section is found, return an empty string.
    if i == len(docstring_lines) - 1:
        return ""

    # Determine the indent level of the example heading;
    # allows example to exist in the middle of docstring.
    heading_indent_level = len(line) - len(line.lstrip())

    # Proceed past the example heading itself.
    i += 1

    # Extract example lines.
    example_lines = []
    for line in docstring_lines[i:]:

        line_indent_level = len(line) - len(line.lstrip())
        if line and line_indent_level <= heading_indent_level:
            # Stop at the next dedent.
            break

        if doc_heading_pattern.match(line):
            # Stop at the next heading.
            break

        example_lines.append(line)

    # Clean up the example string.
    example_string = "\n".join(example_lines)
    example_string = textwrap.dedent(example_string)
    example_string = example_string.strip(" `\n").strip()

    return example_string


def get_kwargs(func: Callable) -> Dict[str, Any]:
    """Get a dictionary of keyword arguments and their default values.

    Args:
        initializer_function: _description_

    Raises:
        ValueError: when the function has arguments with no default value.

    Returns:
        A dictionary of keyword arguments and their default values.
    """
    signature = inspect.signature(func)
    params = signature.parameters
    var_kinds = [inspect.Parameter.VAR_POSITIONAL, inspect.Parameter.VAR_KEYWORD]

    keyword_params = {}
    for k, v in params.items():
        if v.default != inspect.Parameter.empty:
            keyword_params[k] = v.default
        elif v.kind not in var_kinds:
            raise ValueError(
                f"Initializer function '{func.__name__}' has "
                f"an argument '{k}' with no default value."
            )

    return keyword_params


def get_input_docs(func: Callable) -> Dict[str, Optional[str]]:
    """Get a dictionary mapping parameter names to their
    documentation.

    Args:
        func: A function object.

    Returns:
        A dictionary mapping parameter names to their documentation.
    """
    signature = inspect.signature(func)
    signature_param_names = set(signature.parameters)
    arg_start_pattern = re.compile(r"^\s*\**(\w+):\s*(.*)")
    docstring = pydoc.getdoc(func)

    if not docstring:
        warnings.warn(f"No docstring found for function '{func.__name__}'")
        return {param: None for param in signature_param_names}

    docstring_lines = textwrap.dedent(docstring).splitlines()
    current_param_name = None
    params_dict = {}

    for line in docstring_lines:
        stripped_line = line.strip()

        if not stripped_line:
            current_param_name = None

        elif match_ := arg_start_pattern.match(line):
            # Starting to parse new parameter documentation.
            param_name = match_.group(1)

            if param_name not in signature_param_names:
                # Misidentified as parameter or extra parameter in docstring.
                continue

            param_kind = signature.parameters[param_name].kind
            if param_kind == inspect.Parameter.VAR_POSITIONAL:
                param_name = "*" + param_name
            elif param_kind == inspect.Parameter.VAR_KEYWORD:
                param_name = "**" + param_name

            # Record first line in doc for parameter.
            params_dict[param_name] = [match_.group(2)]
            current_param_name = param_name

        elif current_param_name:
            params_dict[current_param_name].append(stripped_line)

    # Check for any missing parameters in the docstring
    missing_params = signature_param_names - params_dict.keys()
    for missing_param in missing_params:
        if signature.parameters[missing_param].kind not in [
            inspect.Parameter.VAR_POSITIONAL,
            inspect.Parameter.VAR_KEYWORD,
        ]:
            warnings.warn(
                f"Parameter '{missing_param}' in function signature of "
                f"'{func.__name__}' is not documented."
            )

    return {k: " ".join(lines) for k, lines in params_dict.items()}


def pick_largest_fs(*dirs) -> Path:
    """Returns the path residing in the largest filesystem.

    If multiple paths are equally large, the first one is returned.
    """
    largest_free_space = -1
    largest_dir = None

    for dir_ in dirs:
        dir_path = Path(dir_).expanduser().absolute()
        try:
            free_space = shutil.disk_usage(dir_path).free
            if free_space > largest_free_space:
                largest_free_space = free_space
                largest_dir = dir_path
        except (FileNotFoundError, PermissionError):
            continue

    if largest_dir is None:
        raise FileNotFoundError(f"No valid directory among {dirs}")

    return largest_dir



# Copyright 2024 Agnostiq Inc.
"""Implementation of the main blueprint object."""

import os
from functools import partial
from typing import Any, Dict, Generic, List, Tuple, Type, TypeVar, Union

import cloudpickle
from covalent._results_manager.result import Result
from covalent._shared_files.context_managers import active_lattice_manager
from covalent._workflow.electron import Electron, electron
from covalent._workflow.lattice import lattice
from covalent_cloud.dispatch_management.interface_functions import get_result
from covalent_cloud.function_serve.deployment import Deployment, get_deployment
from covalent_cloud.service_account_interface.auth_config_manager import get_api_key
from covalent_cloud.shared.schemas.volume import Volume

from covalent_blueprints.blueprints.executor_map import ExecutorMap
from covalent_blueprints.blueprints.inputs import BlueprintInputs
from covalent_blueprints.blueprints.summary import BlueprintSummary
from covalent_blueprints.logger import bp_log
from covalent_blueprints.reader.script import CovalentScript, ScriptType
from covalent_blueprints.reader.utilities import custom_build_graph

API_KEY_ENV_VAR = "CC_API_KEY"

MISSING_API_KEY_MESSAGE = f"""
Covalent Cloud API key is not set.
Please copy your API key from the Covalent Cloud Dashboard and either run:

    >>> from covalent_blueprints import save_api_key
    >>> save_api_key("<you-api-key>")

or set the environment variable in your shell:

    $ export {API_KEY_ENV_VAR}="<your-api-key>".

Note that the environment variable will override any saved API key.
""".strip()


M = TypeVar("M", bound=ExecutorMap)


class CovalentBlueprint(Generic[M]):
    """A runnable blueprint that represents a Covalent workflow or service."""

    executor_map_type: Type = ExecutorMap

    @classmethod
    def create_executor_map(cls, script: CovalentScript) -> M:
        """Create an executor map for this blueprint."""
        return cls.executor_map_type(script)

    def __init__(self, name: str, script: CovalentScript, executor_map: M):
        self._name = name
        self._title = ""
        self._script = script
        self._inputs = BlueprintInputs(script)
        self._executors = executor_map
        self._ids: List[Tuple[str, str]] = []
        self._example = ""
        self._description = ""

        bp_log.info(
            "Initialized blueprint '%s' with script source %s", self.name, script.source
        )

    @property
    def name(self) -> str:
        """Name of the original Covalent script."""
        return self._name.lstrip("_")

    @property
    def title(self) -> str:
        """Displayed title of the blueprint."""
        return self._title or self.name

    @title.setter
    def title(self, title: str) -> None:
        self._title = title

    @property
    def executors(self) -> M:
        """A map from electron/service names to corresponding executors."""
        return self._executors

    @property
    def environments(self) -> List[Dict[str, Any]]:
        """A map from environment names to corresponding environment
        strings."""
        return [
            {k: v for k, v in env.__dict__.items() if k != "settings"}
            for env in self._script.environments
        ]

    @property
    def volume(self) -> Union[Volume, None]:
        """Volume object for the blueprint."""
        return self._script.volume

    @property
    def volumes(self) -> List[Dict[str, Any]]:
        """A map from volume names to corresponding volume strings."""
        omit_keys = ["settings", "vtype"]
        return [
            {k: v for k, v in volume.__dict__.items() if k not in omit_keys}
            for volume in self._script.volumes
        ]

    @property
    def inputs(self) -> BlueprintInputs:
        """Inputs to the blueprint."""
        return self._inputs

    @property
    def ids(self) -> List[Tuple[str, str]]:
        """Dispatch and/or function IDs of workflow results and/or
        deployments."""
        return self._ids.copy()

    @property
    def example(self) -> str:
        """Example usage of the blueprint."""
        return self._example

    @example.setter
    def example(self, example: str) -> None:
        self._example = example

    @property
    def description(self) -> str:
        """Description of the blueprint."""
        return self._description or self._script.module.__doc__ or ""

    @description.setter
    def description(self, description: str) -> None:
        if self._description:
            raise ValueError(
                f"Blueprint '{self.name}' already has a description. "
                "Consider editing the module docstring of the source script."
            )
        self._description = description

    def set_default_inputs(self, **kwargs) -> None:
        """Call this method with any kwargs to set the default inputs for the
        blueprint."""
        if len(kwargs) != 0:
            self.inputs.core_kwargs = kwargs

    def _check_api_key(self) -> None:
        api_key = os.getenv("CC_API_KEY") or get_api_key()
        if not api_key:
            raise ValueError(MISSING_API_KEY_MESSAGE)
        bp_log.info("Found API key: %s", api_key[:4] + "*" * len(api_key[4:]))

    def _rebuild(self) -> None:
        """Rebuild blueprint components by modifying tasks and services in
        place."""

        bp_log.info("Registering pickle by value for module: %s", self._script.source)
        cloudpickle.register_pickle_by_value(self._script.module)

        executors_dict = self._executors.map

        # Electrons
        for electron_name, electron_ in self._script.electrons.items():
            # Create replacement electron.
            new_executor = executors_dict[electron_name]
            new_electron = electron(electron_, executor=new_executor)

            # Update object data; used to update the transport graph at runtime.
            electron_.electron_object = new_electron.electron_object  # type: ignore

        # Services
        for service_name, service_ in self._script.services.items():
            service_.executor = executors_dict[service_name]

        # Lattices
        for captured_lattice_ in self._script.lattices:

            # Replace with new lattice object.
            new_lattice = lattice(**captured_lattice_.get_kwargs())

            if captured_lattice_.lattice_obj.__name__ not in self._script.sublattices:
                bp_log.debug(
                    "Replacing lattice '%s' ('%s') with new lattice",
                    captured_lattice_.lattice_obj,
                    captured_lattice_.lattice_obj.__name__,
                )
                captured_lattice_.lattice_obj = new_lattice
                bp_log.debug(
                    "New lattice '%s' ('%s')",
                    captured_lattice_.lattice_obj,
                    captured_lattice_.lattice_obj.__name__,
                )
            else:
                bp_log.debug(
                    "Overwriting (instead of replacing) lattice"
                    " '%s' ('%s') - is sublattice",
                    captured_lattice_.lattice_obj,
                    captured_lattice_.lattice_obj.__name__,
                )
                captured_lattice_.lattice_obj.__dict__ = new_lattice.__dict__.copy()

            bp_log.debug(
                "Overriding build_graph on lattice object %s ('%s')",
                captured_lattice_.lattice_obj,
                captured_lattice_.lattice_obj.__name__,
            )
            captured_lattice_.lattice_obj.build_graph = partial(  # type: ignore
                custom_build_graph,
                captured_lattice_.lattice_obj,
                self._script.electrons,
            )

    def build(self) -> None:
        """Prepare the blueprint for execution."""
        bp_log.info("Checking for API key...")
        self._check_api_key()  # API key is required for envs and volumes

        bp_log.info("Creating environments for blueprint '%s'", self.name)
        self._script.create_envs()

        bp_log.info("Building blueprint '%s'", self.name)
        self._rebuild()

    def run(
        self, *args, wait_for_result: bool = True, **kwargs
    ) -> Union[str, Result, Deployment]:
        """Run the underlying workflow dispatch or service deployment in the
        cloud. Positional and keyword arguments override the default inputs set
        in the blueprint.

        Args:
            wait_for_result: Wait for the workflow to complete or the deployment
                to reach an active state. Set to False to run asynchronously.
                Defaults to True.

        Returns:
            If wait_for_result is True, the workflow result or deployment client
            is returned. Otherwise, the workflow dispatch ID or a not-yet-active
            deployment client is returned. Use `cc.get_result` or
            `cc.get_deployment` with `wait=True` to retrieve the final result or
            active deployment.
        """
        if args:
            raise ValueError(self._args_not_allowed_error())

        if active_lattice_manager.get_active_lattice() is not None:
            raise RuntimeError(
                "The .run() method is not allowed inside a Covalent lattice. "
                "Please call the blueprint directly instead."
            )

        self.build()
        kwargs = self.inputs.override_core_defaults(kwargs)

        # Execute
        bp_log.debug("Running blueprint '%s' with\nkwargs: %s", self.name, kwargs)
        cloud_executable = self._script.get_cloud_executable(**kwargs)
        handle = cloud_executable(**kwargs)

        # Record the dispatch or deployment ID
        id_ = handle.function_id if isinstance(handle, Deployment) else handle
        self._ids.append((self._script.type.value, id_))

        bp_log.info("Cloud executing blueprint '%s' (ID: %s)", self.name, id_)

        if not wait_for_result:
            return handle

        if self._script.type == ScriptType.DISPATCH:
            return _get_result_with_retries(handle)

        if self._script.type == ScriptType.DEPLOY:
            return get_deployment(handle, wait=True)

        raise ValueError("Unknown type of cloud executable")

    def __call__(self, *args, **kwargs) -> Electron:
        """Run the blueprint as an electron or sub-lattice.

        This method does not run the blueprint in the cloud, unless it
        is used inside a lattice.
        """
        if args:
            raise ValueError(self._args_not_allowed_error())

        if active_lattice_manager.get_active_lattice() is None:
            raise RuntimeError(
                "Calling a blueprint directly is not allowed outside a "
                "Covalent lattice. Please use the .run() method instead."
            )

        self.build()
        kwargs = self.inputs.override_core_defaults(kwargs)
        core_function = self._script.get_core_function()

        bp_log.debug("Calling blueprint '%s' with kwargs: %s", self.name, kwargs)

        if self._script.type == ScriptType.DISPATCH:
            bp_log.debug(
                "Creating sublattice from core function '%s' for blueprint '%s'",
                core_function,
                self.name,
            )
            return electron(core_function)(**kwargs)

        if self._script.type == ScriptType.DEPLOY:
            return core_function(**kwargs)

        raise ValueError("Unknown type of cloud executable")

    def summary(self, get_source: bool = False) -> Dict[str, Any]:
        """Get a summary of the blueprint.

        Args:
            get_source: Include the content of the source file.
            Defaults to False.

        Returns:
            A dictionary with the blueprint summary.
        """
        summary_ = BlueprintSummary(
            name=self.name,
            title=self.title,
            description=self.description,
            type=self._script.type.value,
            inputs=self.inputs.to_dict(),
            executors={k: v.__dict__ for k, v in self.executors.map.items()},
            environments=self.environments,
            volumes=[volume["name"] for volume in self.volumes],
            example=self.example,
            source_file=str(self._script.source),
        )
        if get_source:
            summary_.get_source()

        return summary_.model_dump()

    def _args_not_allowed_error(self):
        """Formats error message for non-empty positional arguments."""
        kwarg_docs = []
        for k, v in self.inputs.docs.items():
            kwarg_docs.append(f"    {k}: {v}")

        kwargs_msg = "\n".join(kwarg_docs)

        return (
            "Blueprint execution calls must not have positional arguments. "
            "Please use keyword arguments instead."
            "\n\n"
            f"Available keyword arguments:\n\n{kwargs_msg}"
            "\n\n"
            "Positional arguments are only valid during blueprint initialization."
        )


def _get_result_with_retries(
    dispatch_id: str, max_loops: int = 5
) -> Union[Result, None]:
    """Get the result of a workflow dispatch, with retries to avoid recursion
    errors."""
    loops = 0
    while loops < max_loops:
        try:
            res = get_result(dispatch_id, wait=True)
            res.result.load()

            return res.result.value

        except RecursionError:
            pass

        finally:
            loops += 1

    raise TimeoutError(
        f"""Result for dispatch is still not available. To continue waiting, run:

    import covalent_cloud as cc

    res = cc.get_result("{dispatch_id}", wait=True)
    res.result.load()
    result = res.result.value

"""
    )



# Copyright 2024 Agnostiq Inc.
""""Tests for blueprint execution settings - inside/outside a lattice."""
import io
from contextlib import redirect_stdout
from pathlib import Path
from unittest.mock import patch

import covalent as ct
import covalent_cloud as cc
import pytest

from covalent_blueprints import get_blueprint
from covalent_blueprints.blueprints.templates import SingleServiceBlueprint

CWD = Path(__file__).parent


class _ForceStop(Exception):
    """Use to prevent actual dispatches in tests."""


def _stop_build(*_, **__):
    raise _ForceStop("This is a test.")


def test_run_inside_lattice_fails():
    """Test that .run() is disallowed inside a lattice."""

    module_path = CWD / "samples/_scripts/_deploy_chatbot.py"
    _cls = SingleServiceBlueprint
    ex = cc.CloudExecutor(validate_environment=False)

    chatbot_bp = get_blueprint(module_path=module_path, _cls=_cls)

    @ct.lattice(executor=ex, workflow_executor=ex)
    def _lattice_1():
        return chatbot_bp.run()  # RuntimeError

    buffer = io.StringIO()

    try:
        with patch.object(chatbot_bp, "build", new=_stop_build):

            with pytest.raises(_ForceStop):
                # Confirm patch, no real dispatch will happen.
                chatbot_bp.build()

            with redirect_stdout(buffer):
                # Outermost patch ensures this will raise _ForceStop
                # if RuntimeError is missed.
                cc.dispatch(_lattice_1)()

    except _ForceStop:
        # Avoid real dispatch if missed RuntimeError.
        pass

    printed_msg = buffer.getvalue()
    expected_msg = "ERROR: The .run() method is not allowed inside a Covalent lattice."

    debug_msg = f"Expected: '{expected_msg}', Got: '{printed_msg}'"
    assert expected_msg in buffer.getvalue(), debug_msg


def test_call_outside_lattice_fails():
    """Test that .__call__() is disallowed outside a lattice."""

    module_path = CWD / "samples/_scripts/_deploy_chatbot.py"
    _cls = SingleServiceBlueprint

    chatbot_bp = get_blueprint(module_path=module_path, _cls=_cls)
    msg = "Calling a blueprint directly is not allowed outside a Covalent lattice."

    try:
        with patch.object(chatbot_bp, "build", new=_stop_build):

            with pytest.raises(_ForceStop):
                # Confirm patch, no real dispatch will happen.
                chatbot_bp.build()

            with pytest.raises(RuntimeError, match=msg):
                # Outermost patch ensures this will raise _ForceStop
                # if RuntimeError is missed.
                chatbot_bp()

    except _ForceStop:
        # Avoid real dispatch if missed RuntimeError.
        pass



# Copyright 2024 Agnostiq Inc.
"""Logger for core blueprints functionality."""

import logging
import time
from pathlib import Path


class LocalTimeFormatter(logging.Formatter):
    """Override the converter to use local time."""

    def converter(self, timestamp):
        """Overridden converter method."""
        return time.localtime(timestamp)


# Define the log file path
BLUEPRINTS_LOGFILE = Path.home() / ".cache/covalent/blueprints/logfile.txt"

# Ensure the log file and its directory exist
if not BLUEPRINTS_LOGFILE.exists():
    BLUEPRINTS_LOGFILE.parent.mkdir(parents=True, exist_ok=True)
    BLUEPRINTS_LOGFILE.touch()

# Define the log format
LOG_FORMAT = (
    "> %(asctime)s [%(levelname)s] - %(pathname)s:%(lineno)d\n\n%(message)s\n\n"
)

# Create a logger
bp_log = logging.getLogger("covalent_blueprints")
bp_log.setLevel(logging.DEBUG)

# Create a file handler
file_handler = logging.FileHandler(BLUEPRINTS_LOGFILE)
file_handler.setLevel(logging.DEBUG)

# Create a formatter with local time
formatter = LocalTimeFormatter(
    fmt=LOG_FORMAT,
    datefmt="%Y-%m-%d %H:%M:%S",
)
file_handler.setFormatter(formatter)

# Add the handler to the logger
bp_log.addHandler(file_handler)


def clear_logs() -> None:
    """Delete the blueprints log file."""
    if not BLUEPRINTS_LOGFILE.exists():
        BLUEPRINTS_LOGFILE.parent.mkdir(parents=True, exist_ok=True)
        BLUEPRINTS_LOGFILE.touch()

    with open(BLUEPRINTS_LOGFILE, "w", encoding="utf-8") as log_file:
        log_file.write("")

    print(f"Erased contents of logs file: {BLUEPRINTS_LOGFILE!s}")


def get_logs_content() -> str:
    """Return the contents of the blueprints log file."""
    with open(BLUEPRINTS_LOGFILE, "r", encoding="utf-8") as log_file:
        return "\n".join([log_file.read(), f"[FILE: {BLUEPRINTS_LOGFILE!s}]"])



# Copyright 2024 Agnostiq Inc.
"""Executor mapping for Covalent blueprints."""

from difflib import get_close_matches
from pprint import pformat
from typing import Dict

from covalent_cloud.cloud_executor.cloud_executor import CloudExecutor

from covalent_blueprints.reader.script import CovalentScript


class ExecutorMap:
    """Custom dict-like object to contain a map of the blueprint's
    executors."""

    def __init__(self, script: CovalentScript):
        self._executor_map = {}

        # Map executors

        for electron_name, electron_ in script.electrons.items():
            electron_ = electron_.electron_object  # type: ignore
            executor_data = electron_.get_metadata("executor_data")  # type: ignore
            if not executor_data:
                raise ValueError(
                    f"Electron '{electron_name}' does not have an executor. "
                    "Please assign a CloudExecutor to this electron."
                )

            executor = CloudExecutor(
                num_cpus=executor_data["attributes"]["num_cpus"],
                memory=executor_data["attributes"]["memory"],
                num_gpus=executor_data["attributes"]["num_gpus"],
                gpu_type=executor_data["attributes"]["gpu_type"],
                env=executor_data["attributes"]["env"],
                time_limit=executor_data["attributes"]["time_limit"],
                volume_id=executor_data["attributes"]["volume_id"],
                validate_environment=False,
            )
            self._executor_map[electron_name] = executor

        for service_name, service in script.services.items():
            service_executor = service.executor
            executor = CloudExecutor(
                num_cpus=service_executor.num_cpus,
                memory=service_executor.memory,
                num_gpus=service_executor.num_gpus,
                gpu_type=service_executor.gpu_type,
                env=service_executor.env,
                time_limit=service_executor.time_limit,
                volume_id=service_executor.volume_id,
                validate_environment=False,
            )
            self._executor_map[service_name] = executor

        self._script = script

    def __getitem__(self, key) -> CloudExecutor:
        if key not in self._executor_map:
            raise self._invalid_name_error(key)

        return ExecutorWrapper(self._executor_map[key])

    def __setitem__(self, key, value) -> None:
        if key not in self._executor_map:
            raise self._invalid_name_error(key)

        if not isinstance(value, CloudExecutor):
            raise ValueError(
                f"Invalid assignment (type {type(value).__name__}), "
                "value must be an instance of cc.CloudExecutor."
            )
        if value.env == "default":
            # Set same environment
            old_executor = self._executor_map[key]
            value.env = old_executor.env
        self._executor_map[key] = value

    def __str__(self):
        display_dict = {}
        for target, executor in self._executor_map.items():
            _executor_dict = executor.__dict__.copy()

            # Omit some keys from the string output
            _executor_dict.pop("validate_environment")
            _executor_dict.pop("settings")
            _executor_dict.pop("volume_id")
            display_dict[target] = _executor_dict

        return pformat(display_dict)

    def _invalid_name_error(self, invalid_key) -> ValueError:

        valid_keys_dict = {}
        for k in self._executor_map:
            valid_keys_dict[f"'{k}'"] = (
                "service" if k in self._script.services else "task"
            )

        spacer = "    "
        valid_keys = "\n".join(
            f"{spacer}{k:>20} ({v})" for k, v in valid_keys_dict.items()
        )

        return ValueError(
            f"Invalid task or service name '{invalid_key}'. "
            f"Valid names are \n\n{valid_keys}\n\n"
            "Please select a valid task or service name."
        )

    @property
    def map(self) -> Dict[str, CloudExecutor]:
        """Returns a copy of the underlying executor map."""
        return self._executor_map.copy()

    def items(self):
        """Returns the items of the underlying executor map."""
        return self._executor_map.items()


class ExecutorWrapper:
    """Wraps a CloudExecutor object to allow for validated attribute
    re- assignment."""

    def __init__(self, executor: CloudExecutor):
        self.__dict__["executor"] = executor

    def __setattr__(self, key, value):

        if not hasattr(self.__dict__["executor"], key):
            valid_attrs = self.__dict__["executor"].__dict__.keys()
            guess = get_close_matches(key, valid_attrs, n=1)
            msg = f"'{key}' is not a valid executor attribute."
            raise AttributeError(msg + f" Did you mean '{guess[0]}'?" if guess else msg)

        new_dict = self.__dict__["executor"].__dict__.copy()
        new_dict.update({key: value, "validate_environment": False})

        if key == "gpu_type" and not value and new_dict.get("num_gpus") != 0:
            new_dict["num_gpus"] = 0
            new_dict["gpu_type"] = ""

        new_executor = CloudExecutor(**new_dict)

        # Update original object with validated attributes.
        self.__dict__["executor"].__dict__.update(**new_executor.__dict__)

    def __getattr__(self, item):

        if item == "__dict__":
            # Avoid recursion error.
            return self.__dict__

        if not hasattr(self.__dict__["executor"], item):
            raise AttributeError(f"Attribute '{item}' is not present in CloudExecutor.")

        return getattr(self.__dict__["executor"], item)

    def __str__(self):
        return (
            self.__dict__["executor"]
            .__str__()
            .replace("CloudExecutor", "ExecutorWrapper")
        )



# Copyright 2024 Agnostiq Inc.
from covalent_cloud.service_account_interface.auth_config_manager import save_api_key
from covalent_cloud.swe_management.secrets_manager import store_secret

from .blueprints.utilities import get_blueprint, register_blueprints_dir
from .decorator import blueprint
from .logger import clear_logs, get_logs_content

__all__ = [
    "blueprint",
    "get_blueprint",
    "save_api_key",
    "store_secret",
    "register_blueprints_dir",
    "clear_logs",
    "get_logs_content",
]
